#include "mboxparser.hpp"

namespace fs = boost::filesystem;

using namespace std;

/* Initializes the new object representing a mail.
 * A mail consists of a header and a body. The header and the body are parsed
 * and initialized implicitly by this constructor.
 *
 * The deque must contain the mail line by line.
 */
Mail::Mail(deque <string> *mail, Configuration *config) {
    parseHeader(mail, config);
}

/* Clears the header and body of this mail.
 */
Mail::~Mail() {
    if(header.size() > 0) {
        header.clear();
    }
    if(body.size() > 0) {
        body.clear();
    }
}

/* Reverts the bitstuffing which mbox-writers perform in order to ensure that
 * only one line per mail starts with 'From '.
 */
void
Mail::unbitstuff(string *textline, Configuration *config) {
    // no unbitstuffing requested?
    if(!config->unbitstuff) {
        return;
    }

    const string from = (config->trailing_space)? "From " : "From";

    for(unsigned int i=0; i < textline->size(); i++) {
        // if the first char is not the unbitstuff_char -> nothing to unbitstuff
        if(i == 0 && (*textline)[i] != config->unbitstuff_char) {
            break;
        }
        // unbitstuff_char exists
        else {
            // no unbitstuff_targets characters?
            if(config->unbitstuff_targets.find_first_of((*textline)[i]) == string::npos) {
                // marker 'From '? -> unbitstuff
                if(textline->size() > (i + from.size() - 1) &&
                        textline->compare(i, from.size(), from) == 0) {
                    textline->erase(0, 1);
                }

                // no 'From '-marker -> nothing to unbitstuff
                break;
            }
        }
    }
}

/* Splits the submitted deque (containing the mail line by line) into two
 * deques, the header and the body.
 * Unbitstuffing is performed implicitly depending on the Configuration.
 */
void
Mail::parseHeader(deque <string> *mail, Configuration *config) {
    bool isHeader = true;
    while(!mail->empty()) {
        string line = mail->front();
        mail->pop_front();
        if(isHeader) {
            // an empty line separates the header from the body
            if(line.empty()) {
                isHeader = false;
            }
            header.push_back(line);
        }
        else { // body
            if(config->unbitstuff) {
                unbitstuff(&line, config);
            }
            body.push_back(line);
        }
    }
}

/* Generates unique file names by appending consecutive numbers to the end of
 * the file names if necessary.
 * It is ensured that the file names won't exceed 31 Characters. So, the
 * file names should be quite portable.
 */
string
Mail::generateUniqueFilename(string subject, const fs::path outdpath, long* outf_counter) {
    string fname = "";

    if(subject.length() > 27)
        subject = subject.substr(0, 27);

    // handle empty subjects specially
    if(subject.length() == 0) {
        subject = "no_subject";
        do {
            fname = subject + NumberToString((*outf_counter)++, 6);
        } while (fs::exists(outdpath / (fname + ".eml")));
    }
    else {
        fname = subject;

        int dup_counter = 1;
        while(fs::exists(outdpath / (fname + ".eml"))) {
            if(subject.length() > 23)
                subject = subject.substr(0, 23);
            fname = subject + NumberToString(dup_counter++, 4);
        }
    }

    return fname;
}

/* Replaces all characters that are not [0-9a-zA-Z] by an underscore and
 * replaces multiple consecutive underscores by a single underscore.
 * This should lead to maximum portability of filenames.
 */
void
Mail::fixFilename(string *filename) {
    bool underscore_found = false;
    for(unsigned int i=0; i < filename->length(); i++) {
        const unsigned char c = (*filename)[i];
        if(isalnum(c)) {
            underscore_found = false;
            continue;
        } else {
            // previous char was an underscore -> erase this underscore
            if(underscore_found) {
                filename->erase(i--, 1);
            } else {
                (*filename)[i] = '_';
                underscore_found = true;
            }
        }
    }


}

/* Writes this mail to an eml file.
 * The destination file name is computed depending on the subject of this mail.
 *
 * The outdpath represents the output directory.
 * The outf_counter is used for consecutive numbers for mails without subject.
 */
void
Mail::store(const fs::path outdpath, long* outf_counter) {
    // retrieve the subject of this mail
    string fname = parseHeaderField("subject:");
    // remove special chars from the file name
    fixFilename(&fname);
    // generate a unique file name
    try {
    fname = generateUniqueFilename(fname, outdpath, outf_counter);
    }
    catch(const fs::filesystem_error& ex) {
        cerr << "Could not generate unique filename: " << ex.what() << endl;
        exit(1);
    }

    // the whole path to the output eml file
    fs::path outfpath = outdpath / (fname + ".eml");
    fs::ofstream outfile(outfpath, fs::ofstream::out);
    assure(outfile, outfpath.string());

    while(!header.empty()) {
        outfile << header.front() << endl;;
        header.pop_front();
    }
    while(!body.empty()) {
        outfile << body.front() << endl;;
        body.pop_front();
    }

    outfile.close();
}

/* converts numbers to strings
 */
template <class TYPE> string
Mail::NumberToString(const TYPE & t, const int fillup) {
    ostringstream os;
    os << setfill('0') << setw(fillup) << t;
    return os.str();
}

/* Parses a header field from this mail.
 *
 * field must be i.e. "subject:"
 */
string
Mail::parseHeaderField(string field) {
    deque<string>::const_iterator iter;
    for(iter = header.begin(); iter != header.end(); ++iter) {
        if(boost::istarts_with(*iter, field))
            return boost::trim_copy(iter->substr(field.length()));
    }
    return "";
}

